from os import environ
from typing import Any, List, Literal, Optional, Type, Dict, Callable

from upstash_redis.commands import (
    Commands,
    PipelineCommands,
    JsonCommands,
    PipelineJsonCommands,
)
from upstash_redis.format import cast_response
from upstash_redis.http import make_headers, SyncHttpClient
from upstash_redis.typing import RESTResultT


class Redis(Commands):
    """
    A Redis client that uses the Upstash REST API.

    Blocking example:

    ```python
    from upstash_redis import Redis

    redis = Redis.from_env()

    redis.set("key", "value")
    redis.get("key")
    ```

    To use the async client, use `upstash_redis.asyncio.Redis`.
    """

    def __init__(
        self,
        url: str,
        token: str,
        rest_encoding: Optional[Literal["base64"]] = "base64",
        rest_retries: int = 1,
        rest_retry_interval: float = 3,  # Seconds.
        allow_telemetry: bool = True,
        read_your_writes: bool = True,
    ):
        """
        Creates a new blocking Redis client.

        :param url: UPSTASH_REDIS_REST_URL in the console
        :param token: UPSTASH_REDIS_REST_TOKEN in the console
        :param rest_encoding: the encoding that can be used by the REST API to parse the response before sending it
        :param rest_retries: how many times an HTTP request will be retried if it fails
        :param rest_retry_interval: how many seconds will be waited between each retry
        :param allow_telemetry: whether anonymous telemetry can be collected
        :param read_your_writes: when enabled, any subsequent commands issued by this client are guaranteed to observe
            the effects of all earlier writes submitted by the same client.
        """
        self._url = url
        self._read_your_writes = read_your_writes
        self._sync_token = ""

        # self._allow_telemetry and self._headers need to stay in this class
        # for compatibility with the ratelimit library.
        self._allow_telemetry = allow_telemetry
        self._headers = make_headers(token, rest_encoding, allow_telemetry)

        self._json = JsonCommands(self)
        self._http = SyncHttpClient(
            encoding=rest_encoding,
            retries=rest_retries,
            retry_interval=rest_retry_interval,
            sync_token_cb=self._update_sync_token,
        )

    @property
    def json(self) -> JsonCommands:
        return self._json

    @classmethod
    def from_env(
        cls,
        rest_encoding: Optional[Literal["base64"]] = "base64",
        rest_retries: int = 1,
        rest_retry_interval: float = 3,
        allow_telemetry: bool = True,
        read_your_writes: bool = True,
    ):
        """
        Load the credentials from environment.

        :param rest_encoding: the encoding that can be used by the REST API to parse the response before sending it
        :param rest_retries: how many times an HTTP request will be retried if it fails
        :param rest_retry_interval: how many seconds will be waited between each retry
        :param allow_telemetry: whether anonymous telemetry can be collected
        :param read_your_writes: when enabled, any subsequent commands issued by this client are guaranteed to observe
            the effects of all earlier writes submitted by the same client.
        """

        return cls(
            environ["UPSTASH_REDIS_REST_URL"],
            environ["UPSTASH_REDIS_REST_TOKEN"],
            rest_encoding,
            rest_retries,
            rest_retry_interval,
            allow_telemetry,
            read_your_writes,
        )

    def __enter__(self) -> "Redis":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Any,
    ) -> None:
        self.close()

    def close(self):
        """
        Closes the resources associated with the client.
        """
        self._http.close()

    def _update_sync_token(self, new_token: str):
        if self._read_your_writes:
            self._sync_token = new_token

    def _maybe_set_sync_token_header(self, headers: Dict[str, str]):
        if self._read_your_writes:
            headers["Upstash-Sync-Token"] = self._sync_token

    def execute(self, command: List) -> RESTResultT:
        """
        Executes the given command.
        """
        self._maybe_set_sync_token_header(self._headers)
        res = self._http.execute(
            url=self._url,
            headers=self._headers,
            command=command,
        )
        return cast_response(command, res)

    def pipeline(self) -> "Pipeline":
        """
        Create a pipeline to send commands in batches
        """
        return Pipeline(
            url=self._url,
            headers=self._headers,
            http=self._http,
            multi_exec="pipeline",
            set_sync_token_header_fn=self._maybe_set_sync_token_header,
        )

    def multi(self) -> "Pipeline":
        """
        Create a pipeline to send commands in batches as a transaction
        """
        return Pipeline(
            url=self._url,
            headers=self._headers,
            http=self._http,
            multi_exec="multi-exec",
            set_sync_token_header_fn=self._maybe_set_sync_token_header,
        )


class Pipeline(PipelineCommands):
    def __init__(
        self,
        url: str,
        headers: Dict[str, str],
        http: SyncHttpClient,
        multi_exec: Literal["multi-exec", "pipeline"],
        set_sync_token_header_fn: Callable[[Dict[str, str]], None],
    ):
        self._url = f"{url}/{multi_exec}"
        self._headers = headers
        self._http = http

        self._json = PipelineJsonCommands(self)
        self._command_stack: List[List[str]] = []

        self._set_sync_token_header_fn = set_sync_token_header_fn

    @property
    def json(self) -> PipelineJsonCommands:
        return self._json

    def execute(self, command: List) -> "Pipeline":
        """
        Adds command to the command stack which will be sent as a batch
        later

        :param command: Command to execute
        """
        self._command_stack.append(command)
        return self

    def exec(self) -> List[RESTResultT]:
        """
        Executes the commands in the pipeline by sending them as a batch
        """
        self._set_sync_token_header_fn(self._headers)

        res: List[RESTResultT] = self._http.execute(  # type: ignore[assignment]
            url=self._url,
            headers=self._headers,
            command=self._command_stack,
            from_pipeline=True,
        )
        response = [
            cast_response(command, response)
            for command, response in zip(self._command_stack, res)
        ]
        self.reset()
        return response

    def reset(self):
        """
        Resets the commands in the pipeline
        """
        self._command_stack = []

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Any,
    ) -> None:
        self.reset()



from json import loads
from typing import Callable, Dict, List, Optional, Tuple, Union

from upstash_redis.typing import RESTResultT
from upstash_redis.utils import GeoSearchResult


def to_dict(res: List, _) -> Dict:
    """
    Convert a list that contains ungrouped pairs as consecutive elements (usually field-value or similar) into a dict.
    """

    return {res[i]: res[i + 1] for i in range(0, len(res), 2)}


def format_geopos(
    res: List[Optional[List[str]]], _
) -> List[Union[Tuple[float, float], None]]:
    return [
        (float(member[0]), float(member[1])) if isinstance(member, List) else None
        for member in res
    ]


def format_geo_search_response(
    res: List[List[Union[str, List[str]]]],
    with_distance: bool,
    with_hash: bool,
    with_coordinates: bool,
) -> List[GeoSearchResult]:
    """
    Format the raw output given by some Geo commands, usually the ones that return properties of members,
    when additional properties are requested.

    They generally return, if requested, in order:
     - the distance (float)
     - the hash (int)
     - the coordinates as:
        - the longitude
        - the latitude

    All represented as strings
    """

    results: List[GeoSearchResult] = []

    # (metin): mypy has trouble narrowing down the types in this function.
    # We can use typing.cast(str, xxx) to force it manually narrow
    # down the types, but I don't like the idea of calling a function
    # that has a runtime cost to just please the type checker. Until
    # there is a better way of doing it, I wil just ignore the errors.

    for member in res:
        result = GeoSearchResult(member[0])  # type: ignore[arg-type]

        if with_distance:
            result.distance = float(member[1])  # type: ignore[arg-type]

            if with_hash:
                result.hash = int(member[2])  # type: ignore[arg-type]

                if with_coordinates:
                    result.longitude = float(member[3][0])  # type: ignore[arg-type]
                    result.latitude = float(member[3][1])  # type: ignore[arg-type]

            elif with_coordinates:
                result.longitude = float(member[2][0])  # type: ignore[arg-type]
                result.latitude = float(member[2][1])  # type: ignore[arg-type]

        elif with_hash:
            result.hash = int(member[1])  # type: ignore[arg-type]

            if with_coordinates:
                result.longitude = float(member[2][0])  # type: ignore[arg-type]
                result.latitude = float(member[2][1])  # type: ignore[arg-type]

        elif with_coordinates:
            result.longitude = float(member[1][0])  # type: ignore[arg-type]
            result.latitude = float(member[1][1])  # type: ignore[arg-type]

        results.append(result)

    return results


def format_time(res: List[str], _) -> Tuple[int, int]:
    return int(res[0]), int(res[1])


def format_sorted_set_response(res: List[str], _) -> List[Tuple[str, float]]:
    """
    Format the raw output given by Sorted Set commands, usually the ones that return the member-score
    pairs of Sorted Sets.
    """
    it = iter(res)
    return list(zip(it, map(float, it)))


def to_optional_float_list(res: List[Optional[str]], _) -> List[Optional[float]]:
    """
    Format a list of strings representing floats or None values.
    """

    return [float(value) if value is not None else None for value in res]


def format_set(res, command):
    options = command[3:]

    if "GET" in options:
        return res
    return res == "OK"


def string_to_json(res, _):
    if res is None:
        return None

    return loads(res)


def to_json_list(res, command):
    return [string_to_json(value, command) for value in res]


def ok_to_bool(res, _):
    return res == "OK"


def to_bool(res, _):
    return bool(res)


def to_bool_list(res, _):
    return list(map(bool, res))


def to_optional_bool_list(res, _):
    return [bool(value) if value is not None else None for value in res]


def to_optional_float(res, _):
    if res is None:
        return None

    return float(res)


def to_float(res, _):
    return float(res)


def format_scan(res, _):
    return int(res[0]), res[1]


def format_hscan(res, _):
    return int(res[0]), to_dict(res[1], None)


def format_zscan(res, _):
    return int(res[0]), format_sorted_set_response(res[1], None)


def format_zscore(res, _):
    return float(res) if res is not None else res


def format_search(res, command):
    withdist = "WITHDIST" in command
    withhash = "WITHHASH" in command
    withcoord = "WITHCOORD" in command
    if withdist or withhash or withcoord:
        return format_geo_search_response(res, withdist, withhash, withcoord)

    return res


def format_hrandfield(res, command):
    with_values = "WITHVALUES" in command
    if with_values:
        return to_dict(res, command)

    return res


def format_zadd(res, command):
    incr = "INCR" in command
    if incr:
        return float(res) if res is not None else res

    return res


def format_sorted_set_response_with_score(res, command):
    with_scores = "WITHSCORES" in command
    if with_scores:
        return format_sorted_set_response(res, command)

    return res


def format_xread_response(res, command):
    """Format XREAD/XREADGROUP response to convert stream entries."""
    if res is None:
        return []
    if not res:
        return res

    # Return as-is to maintain Redis-compatible format
    return res


def format_xrange_response(res, command):
    """Format XRANGE/XREVRANGE response to convert entries."""
    if not res:
        return res

    # Return as-is to maintain Redis-compatible format
    return res


def format_xpending_response(res, command):
    """Format XPENDING response."""
    # Return as-is to maintain Redis-compatible format
    return res


FORMATTERS: Dict[str, Callable] = {
    "COPY": to_bool,
    "EXPIRE": to_bool,
    "EXPIREAT": to_bool,
    "PERSIST": to_bool,
    "PEXPIRE": to_bool,
    "PEXPIREAT": to_bool,
    "RENAME": ok_to_bool,
    "RENAMENX": to_bool,
    "SCAN": format_scan,
    "GEODIST": to_optional_float,
    "GEOPOS": format_geopos,
    "GEORADIUS": format_search,
    "GEORADIUS_RO": format_search,
    "GEORADIUSBYMEMBER": format_search,
    "GEORADIUSBYMEMBER_RO": format_search,
    "GEOSEARCH": format_search,
    "HEXISTS": to_bool,
    "HGETALL": to_dict,
    "HINCRBYFLOAT": to_float,
    "HRANDFIELD": format_hrandfield,
    "HSCAN": format_hscan,
    "HSETNX": to_bool,
    "JSON.ARRPOP": to_json_list,
    "JSON.GET": string_to_json,
    "JSON.MERGE": ok_to_bool,
    "JSON.MGET": to_json_list,
    "JSON.MSET": ok_to_bool,
    "JSON.NUMINCRBY": string_to_json,
    "JSON.NUMMULTBY": string_to_json,
    "JSON.SET": ok_to_bool,
    "JSON.TOGGLE": to_optional_bool_list,
    "PFADD": to_bool,
    "PFMERGE": ok_to_bool,
    "TIME": format_time,
    "SISMEMBER": to_bool,
    "SMISMEMBER": to_bool_list,
    "SMOVE": to_bool,
    "SSCAN": format_scan,
    "ZADD": format_zadd,
    "ZDIFF": format_sorted_set_response_with_score,
    "ZINCRBY": to_float,
    "ZINTER": format_sorted_set_response_with_score,
    "ZMSCORE": to_optional_float_list,
    "ZPOPMAX": format_sorted_set_response,
    "ZPOPMIN": format_sorted_set_response,
    "ZRANDMEMBER": format_sorted_set_response_with_score,
    "ZRANGE": format_sorted_set_response_with_score,
    "ZRANGEBYSCORE": format_sorted_set_response_with_score,
    "ZREVRANGE": format_sorted_set_response_with_score,
    "ZREVRANGEBYSCORE": format_sorted_set_response_with_score,
    "ZSCAN": format_zscan,
    "ZSCORE": format_zscore,
    "ZUNION": format_sorted_set_response_with_score,
    "INCRBYFLOAT": to_float,
    "PUBSUB NUMSUB": to_dict,
    "FLUSHALL": ok_to_bool,
    "FLUSHDB": ok_to_bool,
    "PSETEX": ok_to_bool,
    "SET": format_set,
    "SETEX": ok_to_bool,
    "SETNX": to_bool,
    "MSET": ok_to_bool,
    "MSETNX": to_bool,
    "HMSET": ok_to_bool,
    "LSET": ok_to_bool,
    "SCRIPT FLUSH": ok_to_bool,
    "SCRIPT EXISTS": to_bool_list,
    "XREAD": format_xread_response,
    "XREADGROUP": format_xread_response,
    "XRANGE": format_xrange_response,
    "XREVRANGE": format_xrange_response,
    "XPENDING": format_xpending_response,
    "XGROUP DESTROY": to_bool,
    "XGROUP CREATECONSUMER": to_bool,
}


def cast_response(command: List[str], response: RESTResultT):
    """
    Given a command and its response, casts the response using the `FORMATTERS`
    map

    :param command: Used to determine the formatting to apply
    :param response: Response to format
    """

    # get main command
    main_command = command[0]
    if len(command) > 1 and main_command == "SCRIPT":
        main_command = f"{main_command} {command[1]}"

    # format response
    if main_command in FORMATTERS:
        return FORMATTERS[main_command](response, command)

    return response



import asyncio
import os
import time
from base64 import b64decode
from json import dumps
from platform import python_version
from typing import Any, Dict, List, Literal, Optional, Union, Callable, Type

import httpx

from upstash_redis import __version__
from upstash_redis.errors import UpstashError
from upstash_redis.typing import RESTResultT


def make_headers(
    token: str, encoding: Optional[Literal["base64"]], allow_telemetry: bool
) -> Dict[str, Any]:
    headers = {
        "Authorization": f"Bearer {token}",
    }

    if encoding == "base64":
        headers["Upstash-Encoding"] = encoding

    if allow_telemetry:
        headers["Upstash-Telemetry-Sdk"] = f"py-upstash-redis@v{__version__}"
        headers["Upstash-Telemetry-Runtime"] = f"python@v{python_version()}"

        if os.getenv("VERCEL"):
            platform = "vercel"
        elif os.getenv("AWS_REGION"):
            platform = "aws"
        else:
            platform = "unknown"

        headers["Upstash-Telemetry-Platform"] = platform

    return headers


class SyncHttpClient:
    def __init__(
        self,
        encoding: Optional[Literal["base64"]],
        retries: int,
        retry_interval: float,
        sync_token_cb: Optional[Callable[[str], None]] = None,
    ):
        self._encoding = encoding
        self._retries = retries
        self._retry_interval = retry_interval
        self._sync_token_cb = sync_token_cb
        self._client = httpx.Client(timeout=None)

    def execute(
        self,
        url: str,
        headers: Dict[str, str],
        command: List,
        from_pipeline: bool = False,
    ) -> Union[RESTResultT, List[RESTResultT]]:
        command = _format_command(command, from_pipeline=from_pipeline)

        response: Optional[Dict[str, Any]] = None
        last_error: Optional[Exception] = None

        for attempts_left in range(max(0, self._retries), -1, -1):
            try:
                r = self._client.post(url, headers=headers, json=command)
                sync_token = r.headers.get("Upstash-Sync-Token")
                if self._sync_token_cb and sync_token:
                    self._sync_token_cb(sync_token)

                response = r.json()

                break  # Break the loop as soon as we receive a proper response
            except Exception as e:
                last_error = e

                if attempts_left > 0:
                    time.sleep(self._retry_interval)

        if response is None:
            assert last_error is not None

            # Exhausted all retries, but no response is received
            raise last_error
        if not from_pipeline:
            return format_response(response, self._encoding)
        else:
            return [
                format_response(sub_response, self._encoding)  # type: ignore[arg-type]
                for sub_response in response
            ]

    def close(self):
        self._client.close()

    def __enter__(self) -> "SyncHttpClient":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Any,
    ) -> None:
        self.close()


class AsyncHttpClient:
    def __init__(
        self,
        encoding: Optional[Literal["base64"]],
        retries: int,
        retry_interval: float,
        sync_token_cb: Optional[Callable[[str], None]] = None,
    ):
        self._encoding = encoding
        self._retries = retries
        self._retry_interval = retry_interval
        self._sync_token_cb = sync_token_cb
        self._client = httpx.AsyncClient(timeout=None)

    async def execute(
        self,
        url: str,
        headers: Dict[str, str],
        command: List,
        from_pipeline: bool = False,
    ) -> Union[RESTResultT, List[RESTResultT]]:
        command = _format_command(command, from_pipeline=from_pipeline)

        response: Optional[Union[Dict, List[Dict]]] = None
        last_error: Optional[Exception] = None

        for attempts_left in range(max(0, self._retries), -1, -1):
            try:
                r = await self._client.post(url, headers=headers, json=command)
                sync_token = r.headers.get("Upstash-Sync-Token")

                if self._sync_token_cb and sync_token:
                    self._sync_token_cb(sync_token)

                response = r.json()
                break  # Break the loop as soon as we receive a proper response
            except Exception as e:
                last_error = e

                if attempts_left > 0:
                    await asyncio.sleep(self._retry_interval)

        if response is None:
            assert last_error is not None

            # Exhausted all retries, but no response is received
            raise last_error

        if not from_pipeline:
            return format_response(response, self._encoding)  # type: ignore[arg-type]
        else:
            return [
                format_response(sub_response, self._encoding)
                for sub_response in response
            ]

    async def close(self):
        await self._client.aclose()

    async def __aenter__(self) -> "AsyncHttpClient":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Any,
    ) -> None:
        await self.close()


def format_response(
    response: Dict[str, Any], encoding: Optional[Literal["base64"]]
) -> RESTResultT:
    """
    Raise exception if the response is an error

    Otherwise, decode if an encoding was used
    """
    if response.get("error"):
        raise UpstashError(response["error"])

    result = response["result"]

    if encoding == "base64":
        return decode(result)

    return result


def decode(raw: RESTResultT) -> RESTResultT:
    """
    Decode the response received from the REST API.
    """

    if isinstance(raw, str):
        return "OK" if raw == "OK" else b64decode(raw).decode()
    elif isinstance(raw, int) or raw is None:
        return raw
    elif isinstance(raw, List):
        return [
            # Decode recursively.
            decode(element)
            for element in raw
        ]
    else:
        raise UpstashError(f"Error decoding data for result type {str(type(raw))}")


def _format_command(command: List[Any], from_pipeline: bool = False):
    """
    Format command

    If not from_pipeline, treat the command as a single command. Otherwise,
    treat it as a list of commands
    """
    if from_pipeline:
        return [
            _format_command(command=pipeline_command, from_pipeline=False)
            for pipeline_command in command
        ]

    return [
        element if isinstance(element, (str, int, float)) else dumps(element)
        for element in command
    ]
